/* s4735.c

   Software for controlling the Si4735 based "PC radio
   with USB,  version 2 with 32 kHz xtal" from 
   AK MODL-BUS Computer GmbH 
   
   More info: http://www.elexs.de/SI4735i.html 

   This program is distributed under the GPL, version 2

   Copyright (C) 2011 Tobias Stoeber
   Based on demo program / Delphi code (C) Burkhard Kainka
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#define INFO "20110821-0.1 Si4735 PC Radio with USB V2 Software (based on libftd2xx) by Tobias Stoeber <tobi@to-st.de>"

#include <ctype.h>
#include <ftd2xx.h>
#include <gtk/gtk.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "rds_pty.h"

/* pin defs */
#define PIN_TX (1 << 0)
#define PIN_RX (1 << 1)
#define PIN_RTS (1 << 2)
#define PIN_CTS (1 << 3)
#define PIN_DTR (1 << 4)
#define PIN_DCD (1 << 5)
#define PIN_CDR (1 << 6)
#define PIN_RI (1 << 7)

/* debugging */
/* #define DEBUG 0 */

#ifdef DEBUG
#define OUTPUT_DEBUG_STRING(str) printf(str);
#else
#define OUTPUT_DEBUG_STRING(str) ;
#endif

/* ============================
   .::. SI4735 RADIO PART .::. 
   ============================ */
  
/* global vars */
char Codearray_RT_A[64];
char Codearray_RT_B[64]; 	
char CodearrayRT2[64];
char CodearrayPS[64];
unsigned char FT_Out_Buffer[0x2000];
unsigned char FT_In_Buffer[0x2000];
int Fehler;
int FT_port = 7;
int Buf_adr = 0;
unsigned int R[14]; 
int DataOut[13] = { 0,0,0,0,0,0,0,0,0,0,0,0,0 };
int f_alt;
int f;
int c;
int Code;
double cap;
int rdsaktiv, rdsTimer, Zeit, UTCoffset, Minuten, Stunden;
int TP, TA, MS, PTY, MS;
int autoTune;
int RxVolMuted = 0;
int RTflagAB = 0;
int RToldbit = 0;

int GTcnt[16] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

/* misc start values, frequency, volume etc. */
int startFreq = 8900;		/* default freq  : 89.00 MHz */
int startVolume = 58;		/* default volume: 58 */

int FM_Band_Selected = 0;
int FM_Band_Bottom[3] = { 8800, 7600, 6580 }; 
int FM_Band_Top[3] = { 10800, 9000, 7400 }; 

int FM_Preset_Freq[48];
int FM_Preset_Bank = 0;

/* FTDI vars */
// struct ftdi_context ftdic;
FT_STATUS	ftStatus;
FT_HANDLE	ftHandle;
unsigned char ucMode = 0x00;

// global RDS text vars
char time_str[25] = "** RDS Zeitinfo **";

/* ==========================
      Gtk vars 
   ========================== */

GtkWidget *button_mute;
GtkWidget *button_smeter_update;   
GtkWidget *button_rds;
GtkWidget *entry_freq;
GtkWidget *entry_rds_psn, *entry_rds_pi, *entry_rds_rt_a, *entry_rds_rt_b, *entry_rds_af, *entry_rds_time;
GtkWidget *label_rds_psn, *label_rds_pi, *label_rds_rt_a, *label_rds_rt_b, *label_rds_af, *label_rds_time;
GtkWidget *entry_rds_pty, *label_rds_pty, *entry_rds_ms, *entry_rds_ta, *entry_rds_tp;
GtkWidget *entry_smeter_rssi, *entry_smeter_snr;
GtkWidget *label_smeter_rssi, *label_smeter_snr;
GtkWidget *button_fm_band_norm, *button_fm_band_jp, *button_fm_band_oirt;
GtkWidget *button_fm_preset[6], *button_fm_preset_bank;
GtkWidget *button_fm_autoscan; 
GtkObject *adj_tune, *adj_vol;

GtkTooltips *button_tips;
GtkTooltips *entry_tips;

/* colors for text entry backgrounds M/S, TP, TA */
GdkColor colorWhite = { 0, 65535, 65535, 65535 };
GdkColor colorGreen = { 1, 0, 65535, 0 };

/* ===========================
       some helper functs
   =========================== */
   
/* get hi part */ 
int hi(int d)
{
  int ret = d >> 8;
  return ret;
}

/* get low part */
int lo(int d)
{
  int ret = ( d & 0b0000000011111111 );
  return ret;
}

/* delay in ms */
void Delay(int t)
{
  usleep(t * 1000);
}

int Write_USB_Device_Buffer(DWORD Write_Count)
{
  /* Writes Write_Count Bytes from FT_Out_Buffer to the USB device
   * Function returns the number of bytes actually sent
   * In this example, Write_Count should be 32k bytes max
   */
   
  DWORD Write_Result = 0;
  
  Write_Result = FT_Write(ftHandle, FT_Out_Buffer, Write_Count, &Write_Result);

  return (Write_Result);
}

int Read_USB_Device_Buffer(DWORD Read_Count) 
{
  /* Reads Read_Count Bytes ( or less ) from the USB device to the FT_In_Buffer
   * Function returns the number of bytes actually received  which may range from zero
   * to the actual number of bytes requested, depending on how many have been received
   * at the time of the request + the read timeout value.
   */
   
  DWORD Read_Result = 0;
      
  Read_Result = FT_Read(ftHandle, FT_In_Buffer, Read_Count, &Read_Result);

  return(Read_Result);
}

/* ************************************************
   I2C Routines
   ************************************************ */

void SCL(int d)
{
  if ( d == 0 ) {
    FT_port = FT_port & (255-4);
  } else {
    FT_port = FT_port | (4); /* Bit 2 RTS */
  }

  FT_Out_Buffer[Buf_adr] = FT_port;
  Buf_adr = Buf_adr + 1;
}

void SDA(int d)
{
  if ( d == 0 ) {
    FT_port = FT_port & (255-16);
  } else {
    FT_port = FT_port | (16); /* Bit 4 */
  }

  FT_Out_Buffer[Buf_adr] = FT_port;
  Buf_adr = Buf_adr + 1;
}

void I2C_Init()
{
  Buf_adr = 0;
  SCL(1); SDA(1);          /* SCL=1, SDA=1 */
}

void Start()
{
  SDA(0);          /* SDA=0 */
  SCL(0);          /* SCL=0 */
}

void Stop()
{
  SCL(0); SDA(0);  /* SCL=0, SDA=0 */
  SCL(1);          /* SCL=1 */
  SDA(1);          /* SDA=1 */ 
}

void Acknowledge()
{
  SCL(0); SDA(0);  /* SCL=0, SDA=0 */ 
  SCL(1);          /* SCL=1 */
  SCL(0);          /* SCL=0 */ 
}

void KeinAcknowledge()
{
  SCL(0); SDA(1);  /* SCL=0, SDA=1 */
  SCL(1);          /* SCL=1 */
  SCL(0);          /* SCL=0 */
}

void ChipReset()
{
  OUTPUT_DEBUG_STRING("ChipReset();\n");

  FT_Out_Buffer[0] = 20;
  Buf_adr = 1;
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);
  Delay(10);
  FT_Out_Buffer[0] = 21;
  Buf_adr = 1;
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);
  Delay(10);
}

void Ausgeben(int Wert)
{
  OUTPUT_DEBUG_STRING( "Ausgeben()\n" );

  int Bitwert, n;

  Bitwert = 128;

  for ( n = 1; n <= 8; n++ )
  {
    if ( Bitwert == (Wert & Bitwert) )
    { 
      SDA(1);
    }
    else
    {
      SDA(0);   /* SDA setzen */ 
    }

    SCL(1);     /* SCL=1 */
    SCL(0);     /* SCL=0 */

    Bitwert = Bitwert / 2;
  }

  SDA(1);           /* SDA=1 */
  SCL(1);           /* SCL=1, SDA lesen */
  SCL(0);           /* SCL=0 */ 
}

int Readi2c(int Ack)
{
  OUTPUT_DEBUG_STRING("Readi2c()\n");

  int n;
  int d;

  Buf_adr = 0;
  SDA(1);
  
  for ( n = 1; n <= 8; n++ )
  {
     SCL(1);
     SCL(0);
  }

  SDA(Ack);
  SCL(1);
  SCL(0);
  
  Write_USB_Device_Buffer( Buf_adr);
  Read_USB_Device_Buffer( Buf_adr);

  d = 0;

  for ( n = 1; n <= 8; n++ )
  {
    d = d * 2;

    if ( (FT_In_Buffer[0+2*n] & 8) > 0 )
    {
      d = d + 1;
    }
  }

  return d;
}

void I2Cio(int Nrout, int Nrin)
{
  OUTPUT_DEBUG_STRING( "Begin I2Cio()\n" );

  int n;

  Buf_adr = 0;
  I2C_Init();
  Start();

  Ausgeben(34);

  for ( n = 1; n <= Nrout; n++)
  {
    Ausgeben( DataOut[n] );
  }

  Stop();

  Write_USB_Device_Buffer( Buf_adr);

  Read_USB_Device_Buffer( Buf_adr);

  if ( Nrin > 0 )
  {
    Buf_adr = 0;
    Start();
    
    Ausgeben(35);

    Write_USB_Device_Buffer( Buf_adr);
    Read_USB_Device_Buffer( Buf_adr);

    for ( n = 0; n <= (Nrin-1); n++ )
    {
      R[n] = Readi2c(0);
    }

    R[Nrin] = Readi2c(1);

    Buf_adr = 0;
    Stop();

    Write_USB_Device_Buffer( Buf_adr);
    Read_USB_Device_Buffer( Buf_adr);
  }
}

int GET_INT_STATUS()
{
  int status;
  
  DataOut[1] = 0x14;
  I2Cio (1,1);
  status = Readi2c(1);   /* Nack */

  if ( (status & 128) == 128 )
  { 
    return 1;
  }
  else
  {
    return 0;
  }
}

void waitCTS()
{
  int timeout;
  
  if ( ! GET_INT_STATUS() )
  {
     timeout = 0;
     do {
        timeout = timeout + 1;
     } while ( (! GET_INT_STATUS()) && (! (timeout > 1)));
  }
}

void Prop(int Adr, int Data)
{
  OUTPUT_DEBUG_STRING( "Prop()\n" );

  waitCTS();
  DataOut[1] = 0x12;
  DataOut[2] = 0x00;
  DataOut[3] = hi(Adr);
  DataOut[4] = lo(Adr);
  DataOut[5] = hi(Data);
  DataOut[6] = lo(Data);
  I2Cio(6,1);
}

void Init_RDS()
{
  OUTPUT_DEBUG_STRING( "Init_RDS()\n" );

  Prop(0x1502, 0xA001); /* 0x200?? */

  Delay(10);
  
  Prop(0x1500, 1);

  Delay(10);

  Prop(0x1501, 8);
}

/* function Xtal(); actually not needed anymore */
void Xtal()
{
  OUTPUT_DEBUG_STRING( "Xtal()\n" );

  Prop(0x0202, 125);   /* 4 MHz / 125 = 32 kHz */
  Delay(10);
  Prop(0x0201, 32000);
}

void RX_VOLUME (int L)
{
  OUTPUT_DEBUG_STRING( "RX_Volume()\n" );

  Prop(0x4000, L);
}

void PowerDown()
{
  OUTPUT_DEBUG_STRING( "PowerDown()\n" );

  waitCTS();
  DataOut[1] = 0x11;
  I2Cio(1,0);
}

void GetRev()
{
  OUTPUT_DEBUG_STRING( "GetRev()\n" );

  waitCTS();
  DataOut[1] = 0x10;
  I2Cio(1,8);
  
  int k;
  
  for ( k = 0; k <= 7; k++ )
  {
    printf(" GetRev R[%d] = %d ( %x) \n", k, R[k], R[k]);
  }
}

void Init_AM()
{
  waitCTS();
  DataOut[1] = 0x01;
  DataOut[2] = 0x11;  /* 32 kHz */
  DataOut[3] = 0x05;
  I2Cio(3,0);
  Delay(550);
}

void Init_FM()
{
  waitCTS();
  DataOut[1] = 0x01;
  DataOut[2] = 0x10;   /* 32 kHz */
  DataOut[3] = 0x05;
  I2Cio(3,0);
  Delay(550);
}

void AM_Tune_Freq(int f)
{
  OUTPUT_DEBUG_STRING("AM_Tune_Freq()\n");

  waitCTS();
  DataOut[1] = 0x40;
  DataOut[2] = 0x00;
  DataOut[3] = hi(f);
  DataOut[4] = lo(f);
  DataOut[5] = 0x00;
  DataOut[6] = 0x00;
  I2Cio(6,0);
}

void FM_Tune_Freq(int f)
{
  OUTPUT_DEBUG_STRING("FM_Tune_Freq()\n");

  waitCTS();
  DataOut[1] = 0x20;
  DataOut[2] = 0x00;
  DataOut[3] = hi(f);
  DataOut[4] = lo(f);
  DataOut[5] = 0x00;
  DataOut[6] = 0x00;
  I2Cio(6,0);
}

void AM_Seek_Freq(int Arg1)
{
  OUTPUT_DEBUG_STRING("AM_Seek_Freq()\n");

  waitCTS();
  DataOut[1] = 0x41;
  DataOut[2] = Arg1;
  I2Cio(2,0);
}

void FM_Seek_Freq(int Arg1)
{
  OUTPUT_DEBUG_STRING("FM_Seek_Freq()\n");

  waitCTS();
  DataOut[1] = 0x21;
  DataOut[2] = Arg1;
  I2Cio(2,0);
}

void AM_Tune_Status(int Arg1)
{
  OUTPUT_DEBUG_STRING("AM_Tune_Status()\n");

  waitCTS();
  DataOut[1] = 0x42;
  DataOut[2] = Arg1;
  I2Cio(2,8);
}

void FM_RDS_STATUS()
{
  OUTPUT_DEBUG_STRING("FM_RDS_Status()\n");

  //waitCTS();
  DataOut[1] = 0x24;
  DataOut[2] = 0x07;
  I2Cio(2,13);
}

void AM_RSQ_Status()
{
  OUTPUT_DEBUG_STRING("AM_RSQ_Status()\n");

  waitCTS();
  DataOut[1] = 0x43;
  DataOut[2] = 0x00;
  I2Cio(2,5);
}

void FM_Tune_Status(int Arg1)
{
  OUTPUT_DEBUG_STRING("FM_Tune_Status()\n");

  waitCTS();
  DataOut[1] = 0x22;
  DataOut[2] = Arg1;
  I2Cio(2,8);

/*
  int k;
  
  for ( k = 0; k <= 7; k++ )
  {
    printf(" FM Tune Status R[%d] = %d ( %x) \n", k, R[k], R[k]);
  }
*/

  int f_tuned = (R[2] * 256 + R[3]);
  
  char lt[11];
  sprintf(lt, "%.2f", (double) f_tuned / 100);

  gtk_entry_set_text(GTK_ENTRY(entry_freq), lt);
}

void FM_RSQ_Status()
{
  OUTPUT_DEBUG_STRING( "FM_RSQ_Status()\n" );

  waitCTS();
  DataOut[1] = 0x23;
  DataOut[2] = 0x00;
  I2Cio(2,5);
}

/*
procedure TForm1.Timer1Timer(Sender: TObject);
{
 waitCTS();
 if ( GET_INT_STATUS() )
 {
   //Delay (20);
  AM_Tune_Status(0);
  f := 256* R[2] + R[3];
  if f = f_alt then Timer1.enabled := false;
  f_alt := f;
  if f= 0 then F := 6075;
  Edit1.text := IntToStr (f);
  Application.ProcessMessages;
  c := 256 * R[6] + R[7];
  cap := c * 0.095 + 7;
  c := round (cap);
  Edit5.text := IntToStr (c)+ ' pF' ;
  Delay (50);
  AM_RSQ_Status;
  //Edit2.text := IntToStr (R[1]) + ' ' +   IntToStr (R[2]) +  ' ' +IntToStr (R[3]) +  ' ' +IntToStr (R[4]) +  ' ' +IntToStr (R[5]) +  ' '+IntToStr (R[6])   +  ' '+IntToStr (R[7]) ;
  Edit3.text := IntToStr (R[4]) + ' dBuV';
  Edit4.text := IntToStr (R[5]) + ' dB SNR';
  Delay (20);
 end;
end;
*/

void RDSupdate()
{
  //FILE *fp;
  int n, pos, BLED, altFreq, altFreq2, numFreq, AM;
  int i; // max;
  gchar *RT_utf8;

  altFreq = 0; altFreq2 = 0;
  numFreq = 0;
  AM = 0;
  Minuten = 0;
  Stunden = 0;
  Zeit = 0;

  /* TMC */
  int TMC_T = -1;
  int TMC_F = -1;
  int TMC_DP = -1;
  int TMC_D = -1;
  int TMC_R = -1;
  int TMC_Extent = -1;
  int TMC_Event = -1;
  int TMC_Loc = -1;
  
  for ( n = 1; n <= 1; n++ )
  {
    /* Application.ProcessMessages replaement in Gtk */
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

    if ( GET_INT_STATUS() )
    {
      FM_RDS_STATUS();
    }

    /* open rds logfile for output */
/*
    if ( (fp = fopen("rds.log", "a")) == NULL )
    {
      printf("Cannot open file.\n");
    }
    else
    {
      fprintf(fp, "%c%c%c%c%c%c%c%c%c%c%c%c%c%c\n", R[0],R[1],R[2],R[3], R[4],R[5],R[6],R[7], R[8],R[9],R[10],R[11], R[12],R[13]);
      fclose(fp);
    }
*/
    // printf("RDS - Group Lost?: %d\n", (R[2] & 4)/4);
    // printf("RDS - Num of blocks remaining: %d\n", R[3]);

    if ( R[3] > 0 )
    {
      BLED = R[12] & 3;
 
      if ( (R[6] & 248) == 0 ) 		/* Group 0A - Basic tuning and switching info */
      {
        if ( BLED >= 0 )
        {
          pos = R[7] & 3; 		/* PS address 		*/
          CodearrayPS[pos*2] = R[10]; 	/* high byte 		*/
          CodearrayPS[pos*2+1] = R[11]; /* low byte 		*/
          altFreq = R[8];		/* 1. alternative freq	*/
          altFreq2 = R[9]; 		/* 2. alternative freq	*/
        }

        // OUTPUT_DEBUG_STRING( sprintf("AF debug: %d - %d\n", altFreq, altFreq2) );

        /* 1st alternative frequency */
        if ( altFreq == 250 ) { AM = 1; }	/* AM freq follows */
 
        /* total number of following AFs */
        if ( ( altFreq > 224) && ( altFreq < 250 ) )
        {
          numFreq = altFreq - 224;
          altFreq = 0;
          // OUTPUT_DEBUG_STRING( sprintf("AFs to come: %d\n", numFreq) );
        }
           
        if ( AM == 0 )
        {
          if ( (altFreq > 0) && (altFreq <  205) )
          {
            altFreq = 8750 + altFreq * 10;
          }
        }

        if ( AM == 1 )
        {
          if ( (altFreq > 0) && (altFreq < 16) )
          {
            altFreq = 144 + altFreq * 9;
          }

          if ( (altFreq > 15) && (altFreq < 136) )
          {
            altFreq = 522 + altFreq * 9;
          }
            
          AM = 0;
        }

        /* 2nd alternative frequency */
        if ( altFreq2 == 250 ) { AM = 1; }
 
        if ( AM == 0 )
        {
          if ( (altFreq2 > 0) && (altFreq2 <  205) )
          {
            altFreq2 = 8750 + altFreq2 * 10;
          }
        }

        if ( AM == 1 )
        {
          if ( (altFreq2 > 0) && (altFreq2 < 16) )
          {
            altFreq2 = 144 + altFreq2 * 9;
          }

          if ( (altFreq2 > 15) && (altFreq2 < 136) )
          {
            altFreq2 = 522 + altFreq2 * 9;
          }
            
          AM = 0;
        }
          
        /* TP - Traffic Programme Identification bit (&4) */
        TP = (R[6] & 4) / 4;

        /* TA - Traffic Programme Identification bit (&16) */
        TA = (R[7] & 16) / 16;

        /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
        MS = (R[7] & 8) / 8;
          
        /* PTY - Programm Type Code (&31) */
        int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
        PTY = PTYtmp;
      } 

    else if ( (R[6] & 248) == 1 ) 		/* Group 0B - Basic tuning and switching info */
    {
        if ( BLED >= 0 )
        {
          pos = R[7] & 3; 		/* PS address 		*/
          CodearrayPS[pos*2] = R[10]; 	/* high byte 		*/
          CodearrayPS[pos*2+1] = R[11]; /* low byte 		*/
          /* int PIcode2 = 256 * R[8] + R[9]; */ /* alternative PI code */
        }

        /* TP - Traffic Programme Identification bit (&4) */
        TP = (R[6] & 4) / 4;

        /* TA - Traffic Programme Identification bit (&16) */
        TA = (R[7] & 16) / 16;

        /* M/S - Music / Speech "Switch" bit (&8); 0 = speech | 1 = music or not used */
        MS = (R[7] & 8) / 8;
          
        /* PTY - Programm Type Code (&31) */
        int PTYtmp = ((256 * R[6] + R[7]) >> 5) & 0x1F;
        PTY = PTYtmp;
    } 
    else if ( (R[6] & 248) == 32 ) 	/* Group 2A = Radiotext */
    {
      // OUTPUT_DEBUG_STRING( sprintf("Radiotext n: %d\n", n) );
    
      pos =  R[7] & 15;

      RTflagAB = (R[7] & 16);
      // OUTPUT_DEBUG_STRING( sprintf("RTflagAB: %d\n", RTflagAB ));

      if ( (pos == 0) && (RTflagAB == 0) )
      {
        if ( (R[8] != Codearray_RT_A[0]) && (Codearray_RT_A[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_A[i] = 0x20;
          }
        }
      }

      if ( (pos == 0) && (RTflagAB == 16) )
      {
        if ( (R[8] != Codearray_RT_B[0]) && (Codearray_RT_B[0] != 0x20) )
        {
          for ( i = 0; i <= 63; i++ )
          {
            Codearray_RT_B[i] = 0x20;
          }
        }
      }

      if ( RTflagAB == 0 )
      {
        Codearray_RT_A[pos*4] = R[8];
        Codearray_RT_A[pos*4+1] = R[9];
        Codearray_RT_A[pos*4+2] = R[10];
        Codearray_RT_A[pos*4+3] = R[11];
      }
      else
      {
        Codearray_RT_B[pos*4] = R[8];
        Codearray_RT_B[pos*4+1] = R[9];
        Codearray_RT_B[pos*4+2] = R[10];
        Codearray_RT_B[pos*4+3] = R[11];
      }
    }
    else if ( (R[6] & 248) == 64 ) /* Group 4A - Time and Date */
    {
      UTCoffset = (R[11] & 31) / 2;

      if ( (R[11] & 32) == 32 )
      {
        UTCoffset = UTCoffset * -1;
      }

      Minuten = R[11] / 64 + 4 * (R[10] & 15);
      Stunden = R[10] / 16 + 16 * (R[9] & 1);
      Zeit = 1;

      /* adjust local hrs for UTC+n */ 
      int StundenLocal = Stunden + UTCoffset;
      if ( StundenLocal > 23 )
        StundenLocal = StundenLocal - 24;

      sprintf(time_str, "%02d:%02d / %02d:%02d", Stunden, Minuten, StundenLocal, Minuten);

      // OUTPUT_DEBUG_STRING( sprintf("RDS - 4A - Time: %s\n", time_str) );

      /* update GUI text entry field here, only if new time group found */
      gtk_entry_set_text(GTK_ENTRY(entry_rds_time), (char *)time_str);
    }
    else if ( ((R[6] >> 4) == 8) && ((R[6] & 8) == 0) ) /* Group 8A - TMC */
    {
      OUTPUT_DEBUG_STRING( "RDS - 8A - possible TMC data found!\n" );

      /* T - user message? */
      TMC_T = R[7] & 16;

      /* Single or Multi Group Message? */
      TMC_F = R[7] & 8;

      /* valid? */
      TMC_DP = R[7] & 7;
      
      /* Umfahrung noetig? / redirection? */
      TMC_D = R[8] & 128;
      
      /* direction (positive or negative) */
      TMC_R = R[8] & 64;
      
      /* extent of event in loc codes */
      TMC_Extent = R[8] & 56;

      /* Event code */
      TMC_Event = (R[8] & 7) * 256 + R[9];

      /* Location code */
      TMC_Loc = R[10] * 256 + R[11];
    }
   }      
    int Group1, Group2;
    Group1 = R[6] >> 4;
    Group2 = R[6] & 8;
    //OUTPUT_DEBUG_STRING( sprintf("Group: %d%s\n", Group1, ( Group1 == 0 ) ? "A" : "B") );

    // GTcnt[GTA]++; 
  }

  /* *** update display of RDS values in GUI *** */
  int rds_pi = 256 * R[4] + R[5];
  char pi_str[7];
  sprintf(pi_str, "0x%X", rds_pi);
  char af_str[20];

  if ( (altFreq > 0) && (altFreq2 > 0) )
  {
    sprintf(af_str, "%.2f / %.2f", (double) (altFreq/100), (double) (altFreq2/100));
  }
  else if ( (altFreq > 0) && (altFreq2 == 0) )
  {
    sprintf(af_str, "%.2f", (double) (altFreq/100));
  }
  else if ( (altFreq == 0) && (altFreq2 > 0) )
  {
    sprintf(af_str, "%.2f", (double) (altFreq2/100));
  }
  else
  {
    sprintf(af_str, "%s", "-- AF --");
  }

  char pty_str[13];
  sprintf(pty_str, "%02d %s", PTY, (char *)PTY_Table[PTY]);

  // RDS time update if time found only
  // gtk_entry_set_text(GTK_ENTRY(entry_rds_time), (char *)time_str);
  gtk_entry_set_text(GTK_ENTRY(entry_rds_af), (char *)af_str);
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pi), (char *)pi_str);
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pty), (char *)pty_str);
  gtk_entry_set_text(GTK_ENTRY(entry_rds_psn), (char *)CodearrayPS);

  /* convert radiotext A string to UTF-8 for display with Gtk ?? */
  RT_utf8 = g_convert ( (gchar*)Codearray_RT_A, sizeof(Codearray_RT_A), "ISO-8859-1", "UTF-8", 
                             NULL, NULL, NULL);

  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_a), (char *)RT_utf8);

  /* convert radiotext B string to UTF-8 for display with Gtk ?? */
  RT_utf8 = g_convert ( (gchar*)Codearray_RT_B, sizeof(Codearray_RT_B), "ISO-8859-1", "UTF-8", 
                             NULL, NULL, NULL);

  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_b), (gchar *)RT_utf8);

  /* update indicators for M/S, TP and TA bit settings */
  ( MS ) ? gtk_widget_modify_base(entry_rds_ms, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_ms, GTK_STATE_NORMAL, &colorWhite);
  ( TP ) ? gtk_widget_modify_base(entry_rds_tp, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_tp, GTK_STATE_NORMAL, &colorWhite);
  ( TA ) ? gtk_widget_modify_base(entry_rds_ta, GTK_STATE_NORMAL, &colorGreen) : gtk_widget_modify_base(entry_rds_ta, GTK_STATE_NORMAL, &colorWhite);
}

void SMeter()
{
  if ( GET_INT_STATUS() )
  {
    FM_RSQ_Status();

    char rssi_str[8];
    sprintf(rssi_str, "%d dBuV", R[4]);

    char snr_str[8];
    sprintf(snr_str, "%d dB", R[5]);

    gtk_entry_set_text(GTK_ENTRY(entry_smeter_rssi), (char *)rssi_str);
    gtk_entry_set_text(GTK_ENTRY(entry_smeter_snr), (char *)snr_str);
  }
}


/* ============================
   .::. Start FM/AM Mode .::. 
   ============================ */

void FM_Start(int f, int vol)
{
     PowerDown();
     Delay(50);
     Init_FM();
     Delay(200);
     
     GetRev();

     Delay(200);
     
     Prop(0x1107, 0x0000); 	/* FM_ANTENNA_INPUT */

     Prop(0x1100, 0x0001); 	/* FM Deemphasis */
     
     Prop(0x1403, 0x0001);
     
     Prop(0x1404, 0x0008);
     
     /* Xtal(); */
     RX_VOLUME(vol);

     Delay(200);
     
     Init_RDS();

     Delay(200);

     FM_Tune_Freq(f);

     Delay(200);

     if ( GET_INT_STATUS() )
     {
       FM_Tune_Status(3);
     }
}

/* ============================
   .::. GTK+ WINDOW PART .::. 
   ============================ */

static gboolean time_handler(GtkWidget *widget)
{
  if (widget->window == NULL) return FALSE;

  // gtk_widget_queue_draw(widget);

  RDSupdate();

  return TRUE;
}

void ende_cb(GtkObject *object)
{
  /* close FT Handle */
  FT_Close(ftHandle);

  /* quit Gtk main window */
  gtk_main_quit();
}

void exit_cb(GtkObject *object)
{
  /* leave application and power down radio before */
  PowerDown();
  Delay(50);

  /* close FT Handle */
  FT_Close(ftHandle);

  /* quit Gtk main window */
  gtk_main_quit();
}

void fm_start_cb(GtkObject *object)
{
  FM_Start(startFreq, startVolume);
}

void fm_seek_cb(GtkObject *object, gpointer user_data )
{
  int n, arg;

  if ( strcmp( (char *)user_data, "seek down" ) == 0)
  {
    arg = 0; // seek down, halt at band end (no wrap)
  }
  else
  {
    arg = 8; // seek up, halt at band end (no wrap)
  }

  FM_Seek_Freq(arg);
  
  sleep(1);
  
  if ( GET_INT_STATUS() )
  {
    FM_Tune_Status(3);
  }

  for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
  for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
  for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; }

  TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
}

void fm_autoscan_cb(GtkObject *object )
{
  int f, f_tuned, n, m, oldFreq;

  /* save olf frequency (curently tuned) */
  if ( GET_INT_STATUS() )
  {
    FM_Tune_Status(3);
  }

  f_tuned = (R[2] * 256 + R[3]);
  oldFreq = f_tuned;

  /* begin at lower band end */
  f = FM_Band_Bottom[ FM_Band_Selected ];
  waitCTS();
  FM_Tune_Freq(f);
  Delay (110);

  for ( m = 0; m < 48; m++ )
  {  
    while ( gtk_events_pending() )
    {
      gtk_main_iteration();
    }

    if ( GET_INT_STATUS() )
    {
      FM_Seek_Freq(8);
    }
      
    sleep(3);
  
    if ( GET_INT_STATUS() )
    {
      FM_Tune_Status(3);
    }

    sleep(2);

    f_tuned = (R[2] * 256 + R[3]);

    if ( f_tuned >= (FM_Band_Top[ FM_Band_Selected ] - 10) )
      break;
    
    FM_Preset_Freq[m] = f_tuned;

    if ( m < 6 )
    {
      char pLabel[20];
      sprintf(pLabel, "%d - %.2f", (m+1), (double)f_tuned/100 );
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    
    }
    
    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
    for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; }

    TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
  }

  /* restore freq tuned before autoscan */
  FM_Tune_Freq(oldFreq);
}

void fm_tune_preset_cb(GtkObject *object, gpointer user_data )
{
  int n;

  int preset = (int)user_data;
  int freq = FM_Preset_Freq[ FM_Preset_Bank * 6 + preset ];
  
  if ( (freq > FM_Band_Bottom[ FM_Band_Selected ]) && (freq < FM_Band_Top[ FM_Band_Selected ]) )
  {
    waitCTS();
    FM_Tune_Freq(freq);
    Delay (110);
  
    sleep(1);
  
    if ( GET_INT_STATUS() )
    {
      FM_Tune_Status(3);
    }

    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32;}
    for ( n = 0; n <= 63; n++ ) { CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; }

    TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
  }
}

void fm_preset_change_bank_cb(GtkObject *object )
{
  int f, m;
  int band = FM_Preset_Bank + 1;
  int p_start = 6 * band;
  char pLabel[20];
  
  if ( FM_Preset_Freq[p_start] > 0 )
  {
    FM_Preset_Bank = band;
  }
  else
  {
    FM_Preset_Bank = 0;
  }

  /* label preset memory bank button */
  sprintf(pLabel, "MEM %d", (FM_Preset_Bank + 1) );
  gtk_button_set_label(GTK_BUTTON(button_fm_preset_bank), pLabel);    
 
  for ( m = 0; m < 6; m++ )
  {
    /* get frequency */
    f = FM_Preset_Freq[ FM_Preset_Bank * 6 + m ];

    /* relabel preset button */
    if ( f > 0 )
    {
      sprintf(pLabel, "%d - %.2f", (FM_Preset_Bank * 6 + m+1), (double)f/100 );
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    
    }
    else
    {
      sprintf(pLabel, "%d - ---", (FM_Preset_Bank * 6 + m+1));
      gtk_button_set_label(GTK_BUTTON(button_fm_preset[m]), pLabel);    
    }
  }  
}

void vol_mute_cb(GtkObject *object )
{
  int Arg;
  
  if ( RxVolMuted == 0 ) // not muted -> mute, change button label
  {
    Arg = 3;  
    RxVolMuted = 1;
    gtk_button_set_label(GTK_BUTTON(button_mute), "Un_mute");
  }
  else // muted -> unmute, change label back
  {
    Arg = 0;
    RxVolMuted = 0;
    gtk_button_set_label(GTK_BUTTON(button_mute), "_Mute");
  }

  Prop(0x4001, Arg);
}

void fm_rds_cb(GtkObject *object)
{
  /* g_print("RDS Knopf gedrueckt....\n"); */

  if ( rdsaktiv == 1 )
  {
    g_source_remove( rdsTimer );
    rdsaktiv = 0;
    gtk_button_set_label(GTK_BUTTON(button_rds), "RDS starten");
  }
  else
  {
    /* timers */
    rdsTimer = g_timeout_add(25, (GSourceFunc) time_handler, (gpointer) object);
    rdsaktiv = 1;
    gtk_button_set_label(GTK_BUTTON(button_rds), "RDS stoppen");
  }
}

void fm_smeter_update_cb(GtkObject *object)
{
  /* g_print("S-Meter Knopf gedrueckt....\n"); */

  SMeter();
}

void scrollbar_cb_vol(GtkAdjustment *adj, gpointer user_data)
{
  // print value change
  // g_print("%s hat sich veraendert ",(char *)user_data);
  // g_print("neue Pos. : %f\n", adj->value);

  /* change volume */
  waitCTS();
  RX_VOLUME(adj->value);
}

void change_digits( GtkWidget *widget,
                    GtkSpinButton *spin )
{
  double f = 0.0;

  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (adj_tune),
                              gtk_spin_button_get_value_as_int (spin));

  f = gtk_spin_button_get_value (spin);

  autoTune = 0;

  if ( autoTune == 0 )
  {
    int f2,n;
    f2 = (int) (f * 100);

    waitCTS();
    FM_Tune_Freq(f2);
    Delay (110);

    for ( n = 0; n <= 63; n++ ) { Codearray_RT_A[n] = 32; Codearray_RT_B[n] = 32; CodearrayRT2[n] = 32; }
    for ( n = 0; n <= 32; n++ ) { CodearrayPS[n] = 32; }
  }
}

/* ============================
      .::. MAIN PROGRAM .::. 
   ============================ */

int main(int argc, char **argv)
{
  // struct ftdi_device_list *devlist, *curdev;
  // char manufacturer[128], description[128];

  /* GTK declarations */
  GtkWidget *window;
  GtkWidget *hbox, *vbox;
  GtkWidget *main_vbox;
  GtkWidget *hscrollbar_vol;
  GtkWidget *button_quit, *button_exit;
  GtkWidget *button_fm_start, *button_fm_seek_down, *button_fm_seek_up;
  GtkWidget *frame_tune, *frame_vol, *frame_rds, *frame_btn, *frame_smeter;
  GtkWidget *label;

  // GtkWidget *statusbar;
  // gboolean _isSensitive = FALSE;

  /* FTD2xx declarations */
  int ftPort = 0;
  // DWORD dwBytesInQueue = 0;

  /* print short info line about program */
  printf("%s\n\n", INFO);

  /* Initialize, open (first) device */
  if (FT_Open(ftPort, &ftHandle) != FT_OK)
  {
    printf("Can't open device.\nFTOpen(%d) failed.\n", ftPort);
    puts("Either no device connected or wrong permissions.\nThis may also fail, if ftdi_sio driver is loaded.\nPlease refer to file USAGE for troubleshooting tips.");
    return 1;
  }
   
  /* set sync bitbang mode */
  ftStatus = FT_SetBitMode(ftHandle, PIN_RTS | PIN_DTR, 4);
  if (ftStatus != FT_OK)
  {
    printf("Failed to set bit mode.\n");
    return 1;
  }

  /* set BaudRate */
  ftStatus = FT_SetBaudRate(ftHandle, 153600);  /* Actually 9600 * 16 */
  if (ftStatus != FT_OK)
  {
    printf("Failed to set baud rate.\n");
    return 1;
  }

/*
  // Read out FTDIChip-ID of R type chips
  if (ftdic.type == TYPE_R)
  {
    unsigned int chipid;
    printf("ftdi_read_chipid: %d\n", ftdi_read_chipid(&ftdic, &chipid));
    printf("FTDI chipid: %X\n", chipid);
  }
*/

  /* other declarations */
  rdsaktiv = 0;
  Zeit = 0;
  Minuten = 0;
  Stunden = 0;
  TP = 0; TA = 0; MS = 0; PTY = 0; MS = 0;
  FT_port = 21;        /* 0=TXD,Reset, 2=DTR=SDA, 2=RTS=SCL */

  /* pre-fill DataOut[] */
  int l;
  for ( l = 0; l <= 13; l++ )
  {
    DataOut[l] = 0x0;
  }

  /* pre-fill Codearray / CodearrayPS */
  for ( l = 0; l <= 31; l++ )
  {
    Codearray_RT_A[l] = 0x20;
    Codearray_RT_B[l] = 0x20;
    CodearrayRT2[l] = 0x20;
    CodearrayPS[l] = 0x20;
  }

  for ( l = 32; l <= 63; l++ )
  {
    Codearray_RT_A[l] = 0x20;
    Codearray_RT_B[l] = 0x20;
    CodearrayRT2[l] = 0x20;
  }
  
  /* rest chipset */
  ChipReset();

  Prop(0x1402, 0x0005); 	/* Freq. Spacing 5 kHz */
  Prop(0x1403, 0x0001); 	/* SNR TH */
  Prop(0x1404, 0x0005); 	/* RSSI TH 10 dBuV */
  Prop(0x1501, 4); 		/* RDS FIFO cnt */
  Prop(0x1503, 0x0001); 	/* RDS FIFO cnt */

  /* -------
      GTK+ 
     ------- */

  /* initialise Gtk */
  gtk_init(&argc, &argv);

  /* create new top level window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  /* create main VBox */
  // main_vbox = gtk_vbox_new(FALSE,0);
  main_vbox = gtk_vbox_new (FALSE, 5);
  gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 5);
  gtk_container_add (GTK_CONTAINER (window), main_vbox);

  /* create tooltips */
  button_tips = gtk_tooltips_new();
  entry_tips = gtk_tooltips_new();

  /* create frame for volume setting */
  frame_vol = gtk_frame_new("Lautstärke");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_vol), GTK_SHADOW_ETCHED_IN);

  /* create frame for tuning / freq setting */
  frame_tune = gtk_frame_new("Frequenz / Tuning");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_tune), GTK_SHADOW_ETCHED_IN);

  /* create frame for RDS info etc. */
  frame_rds = gtk_frame_new("RDS Info");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_rds), GTK_SHADOW_ETCHED_IN);

  /* create frame for S-meter info  */
  frame_smeter = gtk_frame_new("S-Meter");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_smeter), GTK_SHADOW_ETCHED_IN);

  /* create frame for buttons */
  frame_btn = gtk_frame_new("Aktionen");
  gtk_frame_set_shadow_type(GTK_FRAME(frame_btn), GTK_SHADOW_ETCHED_IN);

  /* create buttons and add tooltips */
  button_fm_start  = gtk_button_new_with_label("Radio starten");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_start,
                           "Startet Si4735 PC Radio im FM Modus / Initialisierung", "");

  button_fm_seek_down  = gtk_button_new_with_label("- Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_seek_down,
                           "Suchlauf nach unten", "");

  button_fm_seek_up  = gtk_button_new_with_label("+ Suchlauf");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_seek_up,
                           "Suchlauf nach oben", "");

  button_fm_autoscan  = gtk_button_new_with_label("AutoScan");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_autoscan,
                           "Suchlauf uber das gesamte FM Band, belegt Speicherplaetze", "");

  button_fm_band_norm  = gtk_button_new_with_label("FM INT");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_norm,
                           "Setzt Grenzen des FM Bereiches auf 87.5 - 108.0 MHz (EU/US)", "");

  button_fm_band_jp  = gtk_button_new_with_label("FM JP");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_jp,
                           "Setzt Grenzen des FM Bereiches auf 76 - 90 MHz (Japan)", "");
                           
  button_fm_band_oirt  = gtk_button_new_with_label("FM OIRT");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_band_oirt,
                           "Setzt Grenzen des FM Bereiches auf 65.8 - 74.0 MHz (Osteuropa)", "");
                           
  button_rds  = gtk_button_new_with_label("RDS starten");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_rds,
                           "Startet bzw. stoppt den RDS Decoder / die RDS Anzeige", "");

  button_smeter_update  = gtk_button_new_with_label("S-Meter aktualisieren");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_smeter_update,
                           "Aktualisiert die S-Meter Anzeige (RSSI und SNR)", "");

  button_mute  = gtk_toggle_button_new_with_mnemonic("_Mute");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_mute,
                           "Stummschaltung (Mute), Tastenkombi Alt+M", "");

  button_quit  = gtk_button_new_with_label("Radio verlassen");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_quit,
                           "Beendet die Radiosoftware, Radio spielt weiter", "");

  button_exit  = gtk_button_new_with_label("Radio beenden");
  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_exit,
                           "Beendet die Radiosoftware, Radio wird zurueckgesetzt", "");
                           
  /* volume adjustment, RX_VOLUME: 0 <= VOL <= 63, default: 55 */
  adj_vol = gtk_adjustment_new(55.0, 0.0, 63.0, 1.0, 5.0, 1.0);
//  gtk_range_set_update_policy(	GTK_ADJUSTMENT(adj_vol),
//                                GTK_UPDATE_DELAYED); // GTK_UPDATE_DELAYED|GTK_UPDATE_DISCONTINUOUS|GTK_UPDATE_CONTINUOUS
  /* frequency adjustment, FM (87.50 - 108.00), default startFreq */
  adj_tune = gtk_adjustment_new( (float)(startFreq/100), 87.5, 108.0, 0.1, 0.5, 1.0);

  /* horizontal scrollbars for adjustments */
  hscrollbar_vol = gtk_hscale_new(GTK_ADJUSTMENT(adj_vol));

  /* GTk signales */
  gtk_signal_connect_object(GTK_OBJECT(button_fm_start), "clicked",
                           GTK_SIGNAL_FUNC(fm_start_cb), GTK_OBJECT(window));

  gtk_signal_connect(GTK_OBJECT(button_fm_seek_down), "clicked",
                           GTK_SIGNAL_FUNC(fm_seek_cb), "seek down");
  gtk_signal_connect(GTK_OBJECT(button_fm_seek_up), "clicked",
                           GTK_SIGNAL_FUNC(fm_seek_cb), "seek up");

  gtk_signal_connect_object(GTK_OBJECT(button_fm_autoscan), "clicked",
                           GTK_SIGNAL_FUNC(fm_autoscan_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_rds), "clicked",
                           GTK_SIGNAL_FUNC(fm_rds_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_smeter_update), "clicked",
                           GTK_SIGNAL_FUNC(fm_smeter_update_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_mute), "clicked",
                           GTK_SIGNAL_FUNC(vol_mute_cb), GTK_OBJECT(window));

  gtk_signal_connect_object(GTK_OBJECT(button_exit), "clicked",
                           GTK_SIGNAL_FUNC(exit_cb), GTK_OBJECT(window));
  gtk_signal_connect_object(GTK_OBJECT(button_quit), "clicked",
                           GTK_SIGNAL_FUNC(ende_cb), GTK_OBJECT(window));

  gtk_signal_connect(GTK_OBJECT(window), "destroy",
                     GTK_SIGNAL_FUNC(ende_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(adj_vol), "value_changed",
                     GTK_SIGNAL_FUNC(scrollbar_cb_vol), "Scrollbar Volume");

  /* status bar */
  // statusbar = gtk_statusbar_new();

  /* set GTK window parameters  */
  gtk_window_set_title(GTK_WINDOW(window), "Si4735 FM/UKW Radio");	// windows title
  gtk_window_set_default_size(GTK_WINDOW(window), 380, 150);		// default size
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// position
  gtk_container_set_border_width(GTK_CONTAINER(window), 3);		// border

  /* *** pack GUI elements *** */
  //gtk_container_add(GTK_CONTAINER(frame_vol), hscrollbar_vol);

  /* RDS info */
  vbox = gtk_vbox_new(FALSE, 5);

  /* PI */
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_pi = gtk_label_new ("PI: ");
  gtk_label_set_width_chars(GTK_LABEL(label_rds_pi), 4);
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_pi, FALSE, FALSE, 4);
  entry_rds_pi = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pi), "* PI *");
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_pi),8);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_pi), 7);
  gtk_entry_set_has_frame(GTK_ENTRY(entry_rds_pi),TRUE);
  g_object_set(entry_rds_pi, "editable", FALSE, NULL);
  //gtk_widget_set_sensitive(entry_rds_pi, _isSensitive);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_pi, FALSE, FALSE, 4);
  /* PSN */
  label_rds_psn = gtk_label_new ("PSN: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_psn, FALSE, FALSE, 4);
  entry_rds_psn = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_psn), "*RDS PSN*");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_psn), 9);
  g_object_set(entry_rds_psn, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_psn, FALSE, FALSE, 4);
  /* AF */
  label_rds_af = gtk_label_new ("AF: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_af, FALSE, FALSE, 4);
  entry_rds_af = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_af), "*** AF ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_af), 14);
  g_object_set(entry_rds_af, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_af, TRUE, TRUE, 4);
  /* Time */
  label_rds_time = gtk_label_new ("UTC/Loc: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_time, FALSE, FALSE, 4);
  entry_rds_time = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_time), "*** Zeit ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_time), 14);
  g_object_set(entry_rds_time, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_time, TRUE, TRUE, 4);
  /* put hbox in vbox */
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* RadioText part A + PTY*/
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_rt_a = gtk_label_new ("RT A: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_rt_a, FALSE, FALSE, 4);
  entry_rds_rt_a = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_a), "*** RDS Radio Text ***");
  g_object_set(entry_rds_rt_a, "editable", FALSE, NULL);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_rt_a), 63);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_rt_a, FALSE, FALSE, 4);
  label_rds_pty = gtk_label_new ("PTY: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_pty, FALSE, FALSE, 4);
  entry_rds_pty = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_pty), "*** PTY ***");
  g_object_set(entry_rds_pty, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_pty),11);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_pty),11);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_pty, FALSE, FALSE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* RadioText part B + MS + TP + TA */
  hbox = gtk_hbox_new(FALSE, 5);
  label_rds_rt_b = gtk_label_new ("RT B: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_rds_rt_b, FALSE, FALSE, 4);
  entry_rds_rt_b = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_rt_b), "*** RDS Radio Text ***");
  g_object_set(entry_rds_rt_b, "editable", FALSE, NULL);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_rt_b), 63);
  gtk_box_pack_start(GTK_BOX(hbox), entry_rds_rt_b, FALSE, FALSE, 4);
  /* TA */
  entry_rds_ta = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_ta), "TA");
  g_object_set(entry_rds_ta, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_ta),2);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_ta),2);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_ta, FALSE, FALSE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* TP */
  entry_rds_tp = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_tp), "TP");
  g_object_set(entry_rds_tp, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_tp),2);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_tp),2);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_tp, FALSE, FALSE, 4);
  /* MS */
  entry_rds_ms = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_rds_ms), "M/S");
  g_object_set(entry_rds_ms, "editable", FALSE, NULL);
  gtk_entry_set_max_length (GTK_ENTRY(entry_rds_ms),3);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_rds_ms),3);
  gtk_box_pack_end(GTK_BOX(hbox), entry_rds_ms, FALSE, FALSE, 4);

  /* pack RDS boxes */
  gtk_container_border_width (GTK_CONTAINER(frame_rds), 5);
  gtk_container_add(GTK_CONTAINER(frame_rds), vbox);
  gtk_box_pack_start(GTK_BOX(main_vbox), frame_rds, FALSE, FALSE, 0);
  
  /* tuning */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);
  /* frequency */
  label = gtk_label_new ("Freq: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 4);

  entry_freq = gtk_spin_button_new (GTK_ADJUSTMENT(adj_tune), 0.01, 2);

  gtk_tooltips_set_tip (GTK_TOOLTIPS (entry_tips), entry_freq,
                           "Frequenz in MHz, Eingabe direkt, durch Druecken der Pfeile oder mit den Cursortasten bzw. PageUp/PageDown", "");

  gtk_signal_connect (GTK_OBJECT (adj_tune), "value_changed",
                      GTK_SIGNAL_FUNC (change_digits),
                      (gpointer) entry_freq);

  gtk_box_pack_start(GTK_BOX(hbox), entry_freq, FALSE, FALSE, 5);

  label = gtk_label_new ("MHz");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);

  /* seek buttons */
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_seek_down, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_seek_up, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_autoscan, FALSE, FALSE, 2);
  gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_oirt, FALSE, FALSE, 2);
  gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_jp, FALSE, FALSE, 2);
  gtk_box_pack_end(GTK_BOX(hbox), button_fm_band_norm, FALSE, FALSE, 2);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  /* station preset buttons */
  hbox = gtk_hbox_new(FALSE, 5);

  button_fm_preset_bank  = gtk_button_new_with_label( "MEM 1" );

  gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_preset_bank,
                           "Schaltet zwischen den Speicherbaenken mit je 6 Plaetze durch", "");
  
  gtk_signal_connect_object(GTK_OBJECT(button_fm_preset_bank), "clicked",
                            GTK_SIGNAL_FUNC(fm_preset_change_bank_cb), GTK_OBJECT(window));

  gtk_box_pack_start(GTK_BOX(hbox), button_fm_preset_bank, FALSE, FALSE, 2);

  int p;
  for ( p = 0; p < 6; p++ )
  {
    char pLabel[6];
    sprintf(pLabel, "- %d -", p+1);
    button_fm_preset[p]  = gtk_button_new_with_label( pLabel );

    gtk_tooltips_set_tip (GTK_TOOLTIPS (button_tips), button_fm_preset[p],
                             "Speicherplatz", "");

    gtk_signal_connect(GTK_OBJECT(button_fm_preset[p]), "clicked",
                           GTK_SIGNAL_FUNC(fm_tune_preset_cb), (void *)p);

    gtk_box_pack_start(GTK_BOX(hbox), button_fm_preset[p], TRUE, TRUE, 2);
  } 
  
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);

  gtk_container_add(GTK_CONTAINER(frame_tune), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_tune), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox), frame_tune, FALSE, FALSE, 0);

  /* S-meter info */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);
  
  /* smeter update button */
  gtk_box_pack_start(GTK_BOX(hbox), button_smeter_update, FALSE, FALSE, 5);

  /* RSSI */
  label_smeter_rssi = gtk_label_new ("RSSI: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_smeter_rssi, FALSE, FALSE, 4);
  entry_smeter_rssi = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_smeter_rssi), "*** RSSI ***");
  gtk_entry_set_max_length (GTK_ENTRY(entry_smeter_rssi),12);
  gtk_entry_set_has_frame(GTK_ENTRY(entry_smeter_rssi),TRUE);
  gtk_entry_set_width_chars(GTK_ENTRY(entry_smeter_rssi), 10);
  g_object_set(entry_smeter_rssi, "editable", FALSE, NULL);
  gtk_box_pack_start(GTK_BOX(hbox), entry_smeter_rssi, TRUE, TRUE, 4);
  /* SNR */
  label_smeter_snr = gtk_label_new ("SNR: ");
  gtk_box_pack_start(GTK_BOX(hbox), label_smeter_snr, FALSE, FALSE, 4);
  entry_smeter_snr = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry_smeter_snr), "*** SNR ***");
  gtk_entry_set_width_chars(GTK_ENTRY(entry_smeter_snr), 10);
  gtk_box_pack_start(GTK_BOX(hbox), entry_smeter_snr, TRUE, TRUE, 4);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 4);
  /* pack S-meter boxes */
  gtk_container_border_width (GTK_CONTAINER(frame_smeter), 5);
  gtk_container_add(GTK_CONTAINER(frame_smeter), vbox);
  gtk_box_pack_start(GTK_BOX(main_vbox), frame_smeter, FALSE, FALSE, 0);

  /* volume */ 
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), hscrollbar_vol, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_mute, FALSE, FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_vol), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_vol), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox), frame_vol, FALSE, FALSE, 0);

  /* buttons */
  vbox = gtk_vbox_new(FALSE, 5);
  hbox = gtk_hbox_new(TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_fm_start, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_rds, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_quit, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), button_exit, TRUE, TRUE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 5);
  gtk_container_add(GTK_CONTAINER(frame_btn), vbox);
  gtk_container_border_width (GTK_CONTAINER(frame_btn), 5);
  gtk_box_pack_start(GTK_BOX(main_vbox), frame_btn, FALSE, FALSE, 0);

  /* statusbar? */
  // gtk_box_pack_start(GTK_BOX(main_vbox), statusbar, FALSE, TRUE, 1);
  //gtk_container_add(GTK_CONTAINER(window),main_vbox);
  
  /* show Gtk window etc... */
  gtk_widget_show_all(window);

  /* start Gtk main loop */
  gtk_main();

  return EXIT_SUCCESS;
}
